{ FoxClone
  Copyright (C) 2019, 2020 Andy Hardwick
  This source is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.
  This code is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.
  A copy of the GNU General Public License is available on the World Wide Web
  at <http://www.gnu.org/copyleft/gpl.html>. You can also obtain it by writing
  to the Free Software Foundation, Inc., 51 Franklin Street - Fifth Floor,
  Boston, MA 02110-1335, USA.
}

{-------------------------------------------------------------------------------
UNIT BACKUP
Contains backup specific procedures.

-------------------------------------------------------------------------------}
unit backup;

{$mode objfpc}{$H+}

interface

procedure BackBox1(Source : integer);
procedure BackBox2(Source : integer);
procedure BackBox3(Source : integer);
procedure BackBox4(Source : integer);
procedure BackSourceDrive(Source : integer); //populates the source drive check list box
procedure BackConfirm; //displays the confirm backup form
procedure BackBackup;

implementation

uses
  Interfaces, Classes, SysUtils, Forms, Controls, Graphics, Dialogs, ComCtrls,
  CheckLst, StdCtrls, dateutils, process, utilities, progform, mainform, confform;
{-------------------------------------------------------------------------------
PROCEDURE BACKBOX1
Clears and disables checkbox 3 & 4, disables the next button and calls the
procedure to populate checkbox 2.

Parameters:
  Source - the item clicked in the source drive checkbox

Used by: Unit MainFm - BackBox1CK

V1.0 - 06 Jun 2019
V1.1 - 14 Feb 2020 debug code
V1.2 - 20 Mar 2020 checks if the selected drive is the boot drive, if so displays error message
V1.3 - 09 Sep 2020 checks if the selected drive has a loop partition table, displays error message.
                   Reworked the if clauses and introduced error var to make sure that the independent
                   if tests work together.
-------------------------------------------------------------------------------}
procedure BackBox1(Source : integer);
var
  error : boolean = false;
begin
  //debug
  if debug then DebugOut(true,'BackBox1','Source',IntToStr(Source));
  //end debug
  with MainFm do
       begin
         //only one drive can be selected
         BackChkBox1.CheckAll(cbUnchecked);
         //check if the boot drive has been selected
         //if MyDrives[Source + 1].mountpt = BootDrive then
         if IsBoot(MyDrives[Source + 1].mountpt) then
            begin
              BackChkBox2.Enabled:= false;
              BackChkBox2.Clear;
              error:= true;
              MessageDlg('Boot Drive',MyDrives[Source + 1].mountpt + ' cannot be backed up.'
                         + LF + 'It is the foxclone boot drive.',mtError,[mbOK],0)
            end;
         //check if the selected drive has a loop partition table
         if MyDrives[Source + 1].ptabletype = 'LOOP' then
            begin
              BackChkBox2.Enabled:= false;
              BackChkBox2.Clear;
              error:= true;
              MessageDlg('No partition table',MyDrives[Source + 1].mountpt + ' cannot be backed up.'
                         + LF + 'The drive has no partition table.',mtError,[mbOK],0)
            end;
         if not error then
            begin
              BackChkBox1.Checked[Source]:= true;
              //have selected a drive enable the partition box and populate
              BackChkBox2.Enabled:= true;
              BackChkBox2.Clear;
              BackSourceDrive(Source);
            end;
         //clear and disable destination drive list box
         BackChkBox3.Clear;
         BackChkBox3.Enabled:= false;
         BackChkBox3.ScrollWidth:= ChkBoxWidth;
         //Destination partition previously available, clear and disable
         BackChkBox4.Clear;
         BackChkBox4.Enabled:= false;
         BackChkBox4.ScrollWidth:= ChkBoxWidth;
         //if file button previously available, then disable
         NextBtn.Enabled:= false;
       end;
end; //end procedure BackBox1
{-------------------------------------------------------------------------------
PROCEDURE BACKBOX2
Tags the partitions selected for backup in the BackDrive record. Populates
checkbox 3 showing the destination drive for the backup.

Parameters:
  Source - the item clicked in the source drive partition checkbox

Used by: Unit MainFm - BackBox2CK

V1.0 - 06 Jun 2019
V1.1 - 14 Feb 2020 debug code
-------------------------------------------------------------------------------}
procedure BackBox2(Source : integer);
var
  i : integer;
begin
  //debug
  if debug then DebugOut(true,'BackBox2','Source',IntToStr(Source));
  //end debug
  with MainFm do
       begin
         //toggle the item selected
         BackChkBox2.Checked[Source]:= not BackChkBox2.Checked[Source];
         //if select all partitions checked then select them all
         if BackChkBox2.Checked[0] then BackChkBox2.CheckAll(cbChecked);
         //unselect any extended partitions in an MSDOS part table
         for i:= 1 to BackChkBox2.Items.Count - 1 do
             begin
               if BackDrive.parts[i].ptype = 'extended' then BackChkBox2.Checked[i]:= false;
               if BackDrive.parts[i].fs = 'swap' then BackChkBox2.Checked[i]:= false;
               //set the selected partitions in the backdrive record
               BackDrive.parts[i].selected:= BackChkBox2.Checked[i];
             end;
         BackChkBox3.Enabled:= true;
         ShowDrives(BackChkBox3);
         //disable the destination partition check box if enabled
         BackChkBox4.Clear;
         BackChkBox4.Enabled:= false;
         BackChkBox4.ScrollWidth:= ChkBoxWidth;
       end;
end; //end procedure backbox2ck
{-------------------------------------------------------------------------------
PROCEDURE BACKBOX3
Populates the DestDrive record based on the checkbox selection and displays
the partitions in DestDrive in checkbox4.

Parameters:
  Source - the item clicked in the source drive partition checkbox

Used by: Unit MainFm - BackBox3CK

V1.0 - 06 Jun 2019
V1.1 - 14 Feb 2020
-------------------------------------------------------------------------------}
procedure BackBox3(Source : integer);
var
  i : integer;
  maxlength : integer = 0;
  drivestr, partstr : string; //e.g. sda
begin
  //debug
  if debug then DebugOut(true,'BackBox3','Source',IntToStr(Source));
  //end debug
  with MainFm do
       begin
         BackChkBox3.CheckAll(cbUnchecked);
         BackChkBox3.Checked[Source]:= true;
         //have selected a drive enable and clear the partition box
         BackChkBox4.Enabled:= true;
         BackChkBox4.Clear;
         BackChkBox4.ScrollWidth:= ChkBoxWidth;
         //deselect the next button if previously available
         NextBtn.Enabled:= false;
       end;
  //find the selected drive in the array MyDrives. Can't use the itemindex
  //as it will not be showing all drives - source drive missing
  drivestr:= MainFm.BackChkBox3.GetSelectedText;
  drivestr:= trim(copy(drivestr,1,pos(':',drivestr) - 1));
  i:= 1;
  while (i <= TotalDrives) and (MyDrives[i].mountpt <> drivestr) do i+= 1;
  //set the destination drive and add partition info to checkbox4
  DestDrive:= MyDrives[i];
  for i:= 1 to DestDrive.totalparts do
      begin
        partstr:= PartInfo(DestDrive.parts[i],DestDrive.ptabletype);
        if length(partstr) > maxlength then maxlength:= length(partstr);
        MainFm.BackChkBox4.Items.Add(partstr);
      end;
  //set scroll bars
  With MainFm do if maxlength * 6 > ChkBoxWidth then BackChkBox4.ScrollWidth:= ChkBoxWidth + 1
                 else BackChkBox4.ScrollWidth:= ChkBoxWidth;
end; //end procedure backbox3ck
{-------------------------------------------------------------------------------
PROCEDURE BACKBOX4
Checks the partition type - cannot save to unknown filesystems, or extended
partitions. Checks if enough space for the backup on the destination partition.
Enables the save button.

Parameters:
  Source - the item clicked in the destination drive partition checkbox

Used by: Unit MainFm - BackBox4CK

V1.0 - 06 Jun 2019
V1.1 - 14 Feb 2020 debug code
V1.2 - 27 Feb 2020, tidied up code and introduced validparttype and freespace
V1.3 - 27 Apr 2020, checks the filesystem on the destination partition. If fat32 sets split files ON
                    and sets split size to 2048MB. Saves previous settings in oldsettings so they can be
                    restored after backup.
-------------------------------------------------------------------------------}
procedure BackBox4(Source : integer);
var
  i : integer;
  CurrentState : boolean;
begin
  //debug
  if debug then DebugOut(true,'BackBox4','Source',IntToStr(Source));
  //end debug
  CurrentState:= MainFm.BackChkBox4.Checked[Source];
  MainFm.BackChkBox4.CheckAll(cbUnchecked); //only one can be selected
  MainFm.BackChkBox4.Checked[Source]:= not CurrentState; //toggle
  if MainFm.BackChkBox4.Checked[Source] then
     begin
       for i:= 1 to DestDrive.totalparts do DestDrive.parts[i].selected:= false;
       //disable the next/save button
       MainFm.NextBtn.Enabled:= false;
       //check for invalid partition types for destination
       if not ValidPartType(DestDrive.parts[Source+1]) then
          begin
            MainFm.BackChkBox4.Checked[Source]:= false;
            exit;
          end;
       //check if enough space on the destination
       if not FreeSpace(DestDrive.parts[Source+1]) then
          begin
            MainFm.BackChkBox4.Checked[Source]:= false;
            exit;
          end;
       //check filesystem on destination, if fat32 need to switch file splitting on
       if DestDrive.parts[Source+1].fs = 'fat32' then
          begin
            oldsettings:= settings;
            settings.fat32:= true;
            settings.split:= true;
            settings.splitsize:= 2048;
          end;
       MainFm.NextBtn.Enabled:= true;
       MainFm.BackChkBox4.Checked[Source]:= true;
       DestDrive.parts[Source+1].selected:= true; //set the selected partition
     end;
end; //end procedure backbox4
{-------------------------------------------------------------------------------
PROCEDURE BACKSOURCEDRIVE
Gets the drive info in mydrives for the drive selected in the back checkbox 1 and
sets backdrive = drive to backup. Populates back checkbox 2 with the partitions in the selected
drive.

Parameters:
  Source - the item clicked in the source drive checkbox

Used by: BackBox1

V1.0 - 15 Oct 19
V1.1 - 14 Feb 2020 debug code
-------------------------------------------------------------------------------}
procedure BackSourceDrive(Source : integer);
var
  i : Integer;
  partstr : string;
  maxlength : integer = 0;
begin
  //debug
  if debug then DebugOut(true,'BackSourceDrive','Source',IntToStr(Source));
  //end debug
  with MainFm do
     begin
       //set all drives visible except the selected drive in the destination box
       for i:= 1 to TotalDrives do MyDrives[i].show:= true;
       BackDrive:= MyDrives[Source + 1]; //set the drive for backup
       MyDrives[Source + 1].show:= false;
       //display the partitions on the source drive in the SourcePartChkBox
       if BackDrive.totalparts <> 0 then BackChkBox2.Items.Add('Select all partitions');
       for i:= 1 to BackDrive.totalparts do
           begin
             partstr:= PartInfo(BackDrive.parts[i],BackDrive.ptabletype);
             //partstr:= BuildPartStr(
             BackChkBox2.Items.Add(partstr);
             if length(partstr) > maxlength then maxlength:= length(partstr);
           end;
       //set scroll bars on the source partition checklistbox
       if maxlength * 6 > BackChkBox2.Width then BackChkBox2.ScrollWidth:= BackChkBox2.Width + 1
          else BackChkBox2.ScrollWidth:= BackChkBox2.Width;
     end;
end; //end procedure backsourcedrive
{-------------------------------------------------------------------------------
PROCEDURE BACKCONFIRM
The default positions for the objects on the confform are set in the constructor
for the confform and do not need changing for backconfirm (unlike the other
confirm procedures like confirm restore or clone which move the objects around).
Backconfirm populates the confform with the source drive & partitions and
destination file and shows the form.

The key feature is that CurrentAction is set to acback. This is then used by the
event handler for the OK button on the form to determine which procedure to call
to do the backup.

Used by: Unit mainform - TMainFm.BackSaveCK (the save button on the backup tab)

V1.0 - 15 Oct 2019
V1.1 - 14 Feb 2020
-------------------------------------------------------------------------------}
procedure BackConfirm;
var
  i : integer;
  maxlength : integer = 0;
  partstr : string;
begin
  //debug
  if debug then DebugOut(true,'BackConfirm');
  //end debug
  ConfirmForm:= Tconfform.create(MainFm);
  with ConfirmForm do
     begin
       Caption:= 'Confirm Backup';
       //populate the confirm form, first the source drive
       SourceLbl.Caption:= 'Source: ' + BackDrive.model + ' ' + BackDrive.size;
       //now the selected partitions
       PartListBox.Clear;
       for i:= 1 to BackDrive.totalparts do if BackDrive.parts[i].selected then
           begin
             partstr:= PartInfo(BackDrive.parts[i],Backdrive.ptabletype);
             PartListBox.Items.Add(partstr);
             if length(partstr) > maxlength then maxlength:= length(partstr);
           end;
       //set horizontal scroll bar if needed
       if maxlength * 6 > PartListBox.Width then PartListBox.ScrollWidth:= PartListBox.Width + 1;
       //destination drive
       DestLbl.Caption:= 'Destination: ' + DestDrive.model + ' ' + DestDrive.size;
       //backup file
       DestFileLbl.Caption:= 'File: ' + ChopStr(BackFileName + '.backup',40);
       CurrentAction:= acBack;
       ConfirmForm.Show;
     end;
end; //end procedure BackConfirm

{-------------------------------------------------------------------------------
PROCEDURE BACKBACKUP

Does the backup - saves the partition table (sfdisk), grub (on a legacy drive using dd) and then
partclone for each of the selected partitions.

Used by:  Unit confform - Tconfform.OkBtnCK - the ok button on the confirm form

V1.0 - 06 Jun 2019
V1.1 - 21 Oct 19. Added save swap partition info - what partition & UUID
V1.2 - 02 Nov 2019. Problems with clone file to drive because not all parts were
present in the backup - extended and swap excluded. So modify so that the backup
file contains entries for ALL parts with 'excluded'. Code changed to indicate
completion when finished.
V1.3 - 07 Nov 2019 amended to save OS information
V1.4 - 14 Nov 2019 progressfm.free changed to progressfm.close. Was generating errors when backup was
cancelled, foxclone would not quit, terminal error was WARNING: TButton.Destroy with LCLRefCount>0.
Hint: Maybe the component is processing an event?
V1.5 - 23 Nov 2019, quotes added around filenames for grub and sfdisk to cater for spaces in filenames
V1.6 - 25 Nov 2019, backup data set to part table = YES always - a kludge prior to removing the ptable
checkbox on the settings tab.
V1.7 - 07 Jan 2020, grub backup changed to new procedure and part table setting removed from backup file.
V1.8 - 11 Feb 2020, BackBackup was building its own line on the partition info to save to the backup file,
changed so that it uses partinfo instead.
V1.9 - 14 Feb 2020, debug code
V1.10 - 23 Feb 2020, sudo replaced with pkexec and policy added to
                    /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                    to allow sfdisk to run as root.
v1.11 - 27 Apr 2020, at the end checks if settings.fat32 has been set and resets the settings to what
                     they were before the backup - if backup destination is fat32 split files is
                     automatically turned on.
v1.12 - 06 Jun 2020, amended to add foxclone version number and serial number of the drive backed up to
                     the backup file.
v1.13 - 24 Sep 2020, replaced partinfo with writepartinfo, one of the changes consequent on introducing
                     the TRestPartData record in place of partstr which was a simple text str containing
                     partition info.
v1.14 - 15 Jan 2021  testing starttime:= time was returning 31 dec 99, whereas now returned the correct
                     datetime, changed. All occurences of time replaced with now. Incrementing the total
                     blocks moved out of if clause. If clause on cloned successfully, else added so that
                     if not cloned successfully, window says finished.
-------------------------------------------------------------------------------}
procedure BackBackup;
const
  bufsize = 2048;             // Buffer size for reading output
var
  i : integer;
  cmdstr : string;            //string for executing terminal commands
  cmdProcess : TProcess;
  Buffer       : string;
  BytesRead    : integer;
  StartTime,                  //start time for backup
  PstartTime : TdateTime;     //start time for an individual partclone backup
  BackData : TStringList;     //used to store config and then save to file
  tempstr : string;
  PreviousBlocks : Int64 = 0; //blocks backed up in previous partitions
  TotalBlocks : Int64 = 0;    //total blocks to backup
  UmountOK : boolean = false; //is the partition umounted and ok to backup
  complete : string;
begin
  //debug
  if debug then DebugOut(true,'BackBackup');
  //end debug
  ConfirmForm.close; //no longer needed
  MainFm.PageCtrl.Enabled:= false; //disable while running backup
  StartTime:= Now; //start time for partclone backup
  ProgressFm:=TProgressFm.Create(MainFm);
  ProgressFm.Show;
  //How many blocks to backup
  for i:= 1 to BackDrive.totalparts do
      if BackDrive.parts[i].selected then TotalBlocks+= BackDrive.parts[i].pused;
  ProgressFm.OverallProgBar.Max:= TotalBlocks;
  //Is the partition table to be backed up - note it is ALWAYS backed up, this
  //is whether the user has selected the part table for backup. Note it is now redundant.
  if settings.parttable then ProgressFm.PartListBox.Items.Add('Partition table - not started')
  else ProgressFm.PartListBox.Items.Add('Partition table - not selected');
  //populate the list box on the progress form with the # of parts to backup
  for i:= 1 to BackDrive.totalparts do ProgressFm.PartListBox.Items.Add('Part ' + IntToStr(i) + ' - not started');
  if settings.parttable then ProgressFm.Caption:= 'Backing up partition table';
  //save the partition table add quotes around filename to cater for spaces in filename
  cmdstr:= 'pkexec sfdisk --dump /dev/' + BackDrive.mountpt + ' > "' + BackFileName + '.sfdisk"';
  setlength(buffer,bufsize);
  termCmd(cmdstr,cmdProcess); //run sfdisk
  //output from the command is on stderr
  BytesRead:= cmdProcess.stderr.Read(buffer[1],length(buffer));
  application.processmessages;
  if BytesRead > 0 then
     begin
       setlength(buffer,BytesRead);
       MessageDlg('Error saving partition table to ' + BackFileName + '.sfdisk'
                  + LF + buffer,mtError,[mbOK],0);
       if settings.parttable then ProgressFm.PartListBox.Items[0]:= 'Partition table - error';
     end
  else if settings.parttable then ProgressFm.PartListBox.Items[0]:= 'Partition table - complete ';
  cmdProcess.free;
  //backup grub, for a legacy drive the space upto the start of the 1st partition, uses dd
  GrubBackup(BackDrive.mountpt,BackFileName);
  //now run partclone
  for i:= 1 to BackDrive.totalparts do if BackDrive.parts[i].selected then
     begin
       PstartTime:= Now;
       //set current progress bar
       ProgressFm.Caption:= 'Backing up Part ' + IntToStr(i) + ' of ' + IntToStr(BackDrive.totalparts);
       //set the caption in the listbox
        ProgressFm.PartListBox.Items[i]:= 'Part '+ IntToStr(i) + ' - started';
       //build the partclone command line
       cmdstr:= partcloneB(BackDrive.parts[i].pnumber,BackDrive.parts[i].fs);
       writeln('Backup cmd string=',cmdstr);
       //need to umount the partition if mounted
       if CheckMount(BackDrive.parts[i].pnumber) = '<none>' then UmountOK:= true else UmountOK:= false;
       if not UmountOK then if umount(BackDrive.parts[i].pnumber) = 0 then UmountOK:= true;
       if UmountOK then //partition umounted, safe to proceed
          begin
            //run partclone and update progress form
            RunPartClone(cmdstr,BackDrive.parts[i].pused,PreviousBlocks,TotalBlocks, StartTime);
            if CancelPartClone then
               begin
                 CancelPartClone:= false;
                 ProgressFm.Close;
                 exit;
               end;
            PreviousBlocks+= BackDrive.parts[i].pused;
            if ProgressFm.CurrentProgLbl.Caption = 'Cloned successfully' then complete:= ' - complete '
            else complete:= ' - finished ';
            ProgressFm.PartListBox.Items[i]:= 'Part ' + IntToStr(i) + complete
                                              + FormatDateTime('hh:nn:ss',Now - PstartTime);
           end
       else //not safe to backup partition, unable to umount it
          begin
            ProgressFm.PartListBox.Items[i]:= 'Part '+ IntToStr(i) + ' - not backed up';
            MessageDlg('Failed un-mounting ' + BackDrive.parts[i].pnumber + LF +
                       'Not safe to backup a live partition, skipping...',mtError,[mbOK],0)
          end;
     end
  else ProgressFm.PartListBox.Items[i]:= 'Part '+ IntToStr(i) + ' - not selected';

  //now save the configuration
  BackData:= TStringList.Create;
  BackData.Add(FormatDateTime(FDateTimeStr,Now));
  //short term fix - part table checkbox is now irrelevant so always say yes
  //longer term - remove it
  if settings.compress then BackData.Add('Compression:YES')
  else BackData.Add('Compression:NO'); ;
  if settings.split then BackData.Add('Split files:YES')
  else BackData.Add('Split files:NO');
  BackData.Add('Model:' + BackDrive.model);
  BackData.Add('Serial:' + BackDrive.serial); //added in foxclone v40
  BackData.Add('Mount point:' + BackDrive.mountpt);
  for i:= 1 to BackDrive.totalparts do
      if BackDrive.parts[i].selected then with BackDrive do
         begin
           tempstr:= WritePartInfo(parts[i],parts[i].fs);
           //strip out the leading * inserted by partinfo if there is an OS on the partition
           if pos('*',tempstr) = 1 then tempstr:= copy(tempstr,2,99);
           tempstr:= 'Partition:' + tempstr;
           BackData.Add(tempstr);
           tempstr:= 'Blocks:';
           tempstr+= parts[i].pnumber + ',';
           tempstr+= IntToStr(parts[i].pused);
           BackData.Add(tempstr);
          end
      else
        begin
          tempstr:= 'Partition:' + BackDrive.parts[i].pnumber + ' : ';
          if BackDrive.parts[i].ptype = 'extended' then tempstr+= BackDrive.parts[i].ptype
          else tempstr+= BackDrive.parts[i].fs;
          //if a swap partition we need the UUID
          if BackDrive.parts[i].fs = 'swap' then tempstr+= ' : ' + getUUID(BackDrive.parts[i].pnumber);
          BackData.Add(tempstr);
          BackData.Add('Blocks:' + BackDrive.parts[i].pnumber + ',0');
        end;
  BackData.Add('End');
  //don't need quotes around filename to cater for spaces
  BackData.SaveToFile(BackFileName + '.backup');
  BackData.Free;
  //reset the progress form to indicate completion
  ProgressFm.Caption:= 'Backup complete';
  ProgressFm.PartcloneLbl.Caption:= '';
  ProgressFm.CancelBtn.Enabled:= false;
  ProgressFm.FinishBtn.Enabled:= true;
  ProgressFm.AddBtn.Enabled:= true;
  //reset the settings if they have been changed (backup to a fat32 partition)
  if settings.fat32 then settings:= oldsettings;
  //enable the main form
  MainFm.PageCtrl.Enabled:= true;
end; //end procedure BackBackup

end.

