{ FoxClone
  Copyright (C) 2019, 2020 Andy Hardwick
  This source is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.
  This code is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.
  A copy of the GNU General Public License is available on the World Wide Web
  at <http://www.gnu.org/copyleft/gpl.html>. You can also obtain it by writing
  to the Free Software Foundation, Inc., 51 Franklin Street - Fifth Floor,
  Boston, MA 02110-1335, USA.
}

{-------------------------------------------------------------------------------
UNIT CLONE
Contains clone specific procedures and functions for either cloning drive to drive or
cloning from a backup file to a new drive.

-------------------------------------------------------------------------------}
unit clone;

{$mode objfpc}{$H+}

interface
  procedure CloneDDrive(Source : integer);
  procedure CloneDFile(Source : integer);
  procedure CloneDiskConfirm;
  procedure CloneFileConfirm;
  function CloneCheck(Source : string; TargetDrive : string; sfFile : boolean) : boolean;
  procedure CloneDiskToDisk;
  procedure CloneFileOpen(Source : integer);
  procedure CloneFileToDisk;
  procedure CloneLog;


implementation
uses
  Interfaces, Classes, SysUtils, Forms, Controls, Graphics, Dialogs, ComCtrls, CheckLst, StdCtrls,
  dateutils, fileutil, process, strutils, utilities, mainform, progform, confform, noteform;

{-------------------------------------------------------------------------------
PROCEDURE CLONEDDRIVE
Sets the CloneDrive record so it contains the details of the drive to be cloned.

Why using backdrive when drive to drive uses clonedrive and vice versa????

Parameters:
  Source - item index for the drive selected in the clone checkbox1.

Used by: Unit mainform - CloneBox1CK

V1.0 - 15 Oct 2019
V1.1 - 18 Jan 2020, debug
V1.2 - 20 Mar 2020, test added to check if the user has selected the foxclone boot drive.
-------------------------------------------------------------------------------}
procedure CloneDDrive(Source : integer);
var
  i : integer;
begin
  //debug
  if debug then DebugOut(true,'CloneDDrive','Source',IntToStr(Source));
  //end debug
  with MainFm do
    begin
      //Only one drive can be selected
      ClChkBox1.CheckAll(cbUnchecked);
      //check if user has selected the boot drive
      if MyDrives[Source + 1].mountpt <> BootDrive then
         begin
           ClChkBox1.Checked[Source]:= true;
           CloneDrive:= MyDrives[Source + 1]; //set the drive as clone source
           //set all drives visible and selected drive not visible
           for i:= 1 to TotalDrives do MyDrives[i].show:= true;
           MyDrives[Source + 1].show:= false;
           //show drives in checkbox 2 and enable it
           ShowDrives(ClChkBox2);
           ClChkBox2.Enabled:= true;
         end
      else //boot drive has been selected
        begin
          ClChkBox2.Enabled:= false;
          ClChkBox2.Clear;
          MessageDlg('Boot Drive','Cannot clone ' + MyDrives[Source + 1].mountpt + '.'
                     + LF + 'It is the foxclone boot drive.',mtError,[mbOK],0);
        end;
    end;
end; //end procedure CloneDDrive
{-------------------------------------------------------------------------------
PROCEDURE CLONEDFILE
Sets the BackDrive record so it contains the details of the drive to be cloned.

Why using backdrive when drive to drive uses clonedrive and vice versa????
and could have one not two cloneddrive and clonedfile - combine

Parameters:
  Source - item index for the drive selected in the clone checkbox1.

Used by: Unit mainform - CloneBox1CK

V1.0 - 15 Oct 2019
V1.1 - 18 Jan 2020, debug
-------------------------------------------------------------------------------}
procedure CloneDFile(Source : integer);
var
  i : integer;
  partstr : string;
  maxlength : integer = 0;
begin
  //debug
  if debug then DebugOut(true,'CloneDFile','Source',IntToStr(Source));
  //end debug
  with MainFm do
    begin
      ClChkBox2.Clear;
      ClChkBox2.Enabled:= true;
      ClListBox3.Enabled:= false;
      ClListBox3.Clear;
      ClListBox3.ScrollWidth:= ChkBoxWidth;
      ClChkBox4.Enabled:= false;
      ClChkBox4.Clear;
      CloneBtn.Enabled:= false;
      //set all drives visible
      for i:= 1 to TotalDrives do MyDrives[i].show:= true;
      ClChkBox1.CheckAll(cbUnchecked);
      ClChkBox1.Checked[Source]:= true;
      BackDrive:= MyDrives[Source + 1]; //set the drive for the clone source
      for i:= 1 to BackDrive.totalparts do
          begin
            partstr:= PartInfo(BackDrive.parts[i],BackDrive.ptabletype);
            ClChkBox2.Items.Add(partstr);
            if length(partstr) > maxlength then maxlength:= length(partstr);
          end;
      //set scroll bars on the source partition checklistbox
      if maxlength * 6 > ChkBoxWidth then ClChkBox2.ScrollWidth:= ChkBoxWidth + 1
         else ClChkBox2.ScrollWidth:= ChkBoxWidth;
    end;

end; //end procedure CloneDFile
{-------------------------------------------------------------------------------
FUNCTION CLONECHECK
Checks that the target drive is big enough to take the contents of the source
drive. It does this by either generating a sfdisk file or reading the file for
the source and finding the last sector on the drive. The size of the target is
found by calling drivesize which uses blockdev.
Size is measured in 512 byte sectors.
Returns true if the target is big enough for the clone, false otherwise.

Parameters:
  Source - a string defining the source drive, e.g. sda, or the filename for
           a partition table file generated by sfdisk
  TargetDrive - a string defining the target, i.e. destination for the clone,
                in the form sdx.
  sfFile - boolean, defines whether source is a drive or a file, when true
           then source = filename, false then source = drive.

Used by:
  Unit mainform - CloneBox2CK
  Unit mainform - CloneBox4CK

V1.0 - 15 Oct 19
V1.1 - 18 Jan 20, debug
V1.2 - 23 Feb 2020, sudo replaced with pkexec and policy added to
                    /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                    to allow sfdisk to run as root.
V1.3 - 11 May 2020, additional debug code, user reporting not enough space cloning from backup of a
                    320GB drive to a 2TB drive. Problem found, was using strtoint instead of strtoint64,
                    net result was a target sector size that was negative, code modified in clonecheck
                    and in drivesize.
-------------------------------------------------------------------------------}
function CloneCheck(Source : string; TargetDrive : string; sfFile : boolean) : boolean;
const
  bufsize = 2048;
var
  cmdstr : string;
  Buffer : array[1..bufsize] of char;
   cmdProcess : TProcess;
  TargetSectors : int64; //# of sectors on the target drive
  PTable : TStringList;
  cmdStream : TMemoryStream;
  StartStr, FinishStr : string;
  Start : int64 = 0;
  Finish : int64 = 0;
  BytesRead : integer;
  tempstr : string;
begin
  //debug
  if debug then
     begin
       if sfFile then tempstr:= 'true' else tempstr:= 'false';
       DebugOut(false,'CloneCheck','Source',Source,'TargetDrive',TargetDrive,'sfFile',tempstr);
     end;
  //end debug
  //get the sectors on the target drive
  TargetSectors:= DriveSize(TargetDrive);
  PTable:= TStringList.Create;
  if TargetSectors = 0 then
     begin
       MessageDlg('Error with target drive','Failed to get size of drive ' + TargetDrive
                  + LF + 'Clone cannot continue', mtError,[mbOK],0);
       CloneCheck:= false;
     end
  else //get partition table for source drive and end of last sector
    begin
      if sfFile then PTable.LoadFromFile(Source) else
         begin
           cmdstr:= 'pkexec sfdisk --dump /dev/' + Source;
           cmdStream:= TMemoryStream.Create;
           termCmd(cmdstr,cmdProcess);
           //output from cmdProcess is read in a loop until no more data is available
           repeat //should get the output in one hit
             BytesRead:= cmdProcess.Output.Read(Buffer, bufsize);
             cmdStream.Write(Buffer, BytesRead);
           until BytesRead = 0;  // stop if no more data is available
           cmdProcess.Free; //finished with it
           cmdStream.Position:= 0; //make sure we start at the beginning of the stream
           PTable.LoadFromStream(cmdStream);
           cmdStream.Free;
         end;
      //last line in file is the last partition
      tempstr:= trim(copy(PTable[PTable.Count-1],pos('start=',PTable[PTable.Count-1])+6,99));
      Startstr:= copy(tempstr,1,pos(',',tempstr)-1);
      Start:= StrToInt64(StartStr);
      FinishStr:= trim(copy(tempstr,pos(', size=',tempstr)+7,99));
      FinishStr:= copy(FinishStr,1,pos(',',FinishStr)-1);
      Finish:= StrToInt64(FinishStr);
      Finish:= Start + Finish;
      Ptable.Free;
      if debug then
         begin
           writeln('last sector on source, ',Source,' =',inttostr(Finish));
           writeln('sectors on target, ',TargetDrive,' =', inttostr(TargetSectors));
         end;
      if Finish > TargetSectors then
         begin
           MessageDlg('Target drive too small','Source last partition ends at '
                   + FloatToStrF(Finish/2/1024/1024,ffFixed,4,2) + 'GB' + LF
                   + 'Target drive ' + FloatToStrF(TargetSectors/2/1024/1024,ffFixed,4,2) + 'GB' + LF
                   + 'Shrink the source partitions and try again', mtError,[mbOK],0);
           CloneCheck:= false;
         end
      else CloneCheck:= true;
    end;
end; //end procedure CloneCheck
{-------------------------------------------------------------------------------
PROCEDURE CLONEDISKCONFIRM
CloneDiskConfirm populates the confform with the source drive & partitions and
destination file and shows the form. The position of some object are changed
from the defaults.
The key feature is that CurrentAction is set to acClone. This is then used by the
event handler for the OK button on the form to determine which procedure to call
to do the clone.

Used by:
  Unit mainform - CloneBtnCK

V1.0 - 15 Oct 19
V1.1 - 18 Jan 20, debug
-------------------------------------------------------------------------------}
procedure CloneDiskConfirm;
var
  i : integer;
  maxlength : integer = 0;
  Source : string;
begin
  //debug
  if debug then DebugOut(true,'CloneDiskConfirm');
  //end debug
  //this is for drive to drive
  //first confirmation box
  ConfirmForm:= Tconfform.create(MainFm);
  with ConfirmForm do
     begin
       //populate the confirm form
       Caption:= 'Confirm Clone Drive to Drive';
       with CloneDrive do Source:= model + ' : ' + size + ' : ' + 'Type: ' + ptabletype;
       SourceLbl.Caption:= Source;
       DestLbl.Visible:= true;
       DestLbl.Caption:= 'WARNING - the target drive will be overwritten!';
       DestLbl.Font.Color:= ClRed;
       for i:= 0 to MainFm.ClChkBox2.Count - 1 do
           if MainFm.ClChkBox2.Checked[i] then DestFileLbl.Caption:= MainFm.ClChkBox2.Items[i];
       PartListBox.Clear;
       with CloneDrive do
            for i:= 1 to totalparts do PartListBox.Items.Add(PartInfo(parts[i],parts[i].ptype));
       cDriveImg.Picture.LoadFromFile(DriveImgFile);
       cFileImg.Picture.LoadFromFile(DriveImgFile);
       cArrowImg.top:= 65;
       cArrowImg.left:= 30;
       for i:= 0 to PartListBox.Items.Count - 1 do
           if length(PartListBox.Items[i]) > maxlength then
              maxlength:= length(PartListBox.Items[i]);
       if maxlength * 6 > PartListBox.Width then
          PartListBox.ScrollWidth:= PartListBox.Width + 1;
       CurrentAction:= acClone;
       Show;
     end;
end; //end procedure CloneDiskConfirm
{-------------------------------------------------------------------------------
PROCEDURE CLONEFILECONFIRM
CloneFileConfirm populates the confform with the source drive & partitions and
destination file and shows the form. The position of some object are changed
from the defaults.
The key feature is that CurrentAction is set to acFileClone. This is then used
by the event handler for the OK button on the form to determine which procedure
to call to do the clone.

Used by:
  Unit mainform - CloneBtnCK

V1.0 - 15 Oct 2019
V1.1 - 08 Nov 2019, if swap partition present, swap UUID stripped out of the item description
V1.2 - 18 Jan 2020, debug
V1.3 - 23 Sep 2020, minor changes resulting from defining TRestPartData record
-------------------------------------------------------------------------------}
procedure CloneFileConfirm;
var
  i : integer;
  maxlength : integer = 0;
  //Source : string;
begin
  //debug
  if debug then DebugOut(true,'CloneFileConfirm');
  //end debug
  //this is for file to drive
  //first confirmation box
  ConfirmForm:= Tconfform.create(MainFm);
  with ConfirmForm do
     begin
       //populate the confirm form
       Caption:= 'Confirm Clone File to Disk';
       SourceLbl.Caption:= 'File: ' + ChopStr(RestFileName,40);
       DestLbl.Visible:= true;
       DestLbl.Caption:= 'WARNING - the target drive will be overwritten!';
       DestLbl.Font.Color:= ClRed;
       for i:= 0 to MainFm.ClChkBox4.Count - 1 do
           if MainFm.ClChkBox4.Checked[i] then DestFileLbl.Caption:= MainFm.ClChkBox4.Items[i];
       PartListBox.Clear;
       for i:= 1 to RSettings.TotalParts do
            begin
               PartListBox.Items.Add(MainFm.ClListBox3.Items[i-1]);
              //strip out the UUID from any item with a swap partition
              //index on i-1 cos PartListBox indexed from zero
               if pos(': swap :',PartListBox.Items[i-1]) > 0 then with PartListBox do
                  Items[i-1]:= copy(Items[i-1],1,pos(': swap :',Items[i-1]) + 6);
              if length(PartListBox.Items[i-1]) > maxlength then maxlength:= length(PartListBox.Items[i-1]);
            end;
       //set horizontal scroll bar if needed
       if maxlength * 6 > PartListBox.Width then PartListBox.ScrollWidth:= PartListBox.Width + 1;
       cDriveImg.Picture.LoadFromFile(FileImgFile);
       cDriveImg.left:= 30;
       cFileImg.Picture.LoadFromFile(DriveImgFile);
       cArrowImg.top:= 70;
       cArrowImg.left:= 30;
       for i:= 0 to PartListBox.Items.Count - 1 do
           if length(PartListBox.Items[i]) > maxlength then
              maxlength:= length(PartListBox.Items[i]);
       if maxlength * 6 > PartListBox.Width then
          PartListBox.ScrollWidth:= PartListBox.Width + 1;
       CurrentAction:= acFileClone;
       Show;
     end;
end; //end procedure CloneFileConfirm
{-------------------------------------------------------------------------------
PROCEDURE CLONEDISKTODISK
Does the disk to disk clone

V1.1 - added functionality to create a swap partition
V1.2 - changed progress form on completion to indicate that it has finished. Was having problems
cloning a dual boot win + linux drive to a msdos drive with a single ntfs partition. First part on
source was a ms system reserved 579MB ntfs part. This was showing an error in gparted - couldn't read
partition. Doing a reboot and all was okay. Added partprobe at end of procedure to force the kernel to
reread the partition table. This seems to have fixed the problem without having to do a reboot.

TBD replace partprobe with blockdev - one less package for the iso.

Used by: Unit confform - Tconfform.OkBtnCK

V1.0 - 15 Oct 2019
V1.1 - 30 Oct 2019
V1.2 - 01 Nov 2019
V1.3 - 18 Jan 2020, debug
V1.4 - 22 Feb 2020, sudo replaced with pkexec and policy added to
                    /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                    to allow sfdisk and partprobe to run as root.
V1.5 - 05 Mar 2020, bug - was setting the destination partition wrong on an NVME drive or SD card
                    instead of mmcblk0p1 it was using mmcblk01. fixed.
V1.6 - 21 Mar 2020, modified to allow more than 9 partitions on a drive. Was using just the last char
                    of the pnumber, e.g. sda1, now checks 2nd last char.
V1.7 - 15 Jan 2021  testing starttime:= time was returning 31 dec 99, whereas now returned the correct
                    datetime, changed, all occurences of time replaced with now. Log button added to
                    procedure.
V1.8 - 02 Feb 2021  bug fix, when cloning from nvme to sata it was leaving in the p, e.g. trying to
                    clone to sdap1 not sda1
-------------------------------------------------------------------------------}
procedure CloneDiskToDisk;
const
  bufsize = 2048; // Buffer size for reading output
  numbers = '123456789';
  separator = '-----------------------------------------------------------------------------';
var
  StartTime, PstartTime : TDateTime;
  TotalBlocks : int64 = 0;
  i, j : integer;
  cmdStr : string;
  Destination : string; //is this sdx??
  cmdProcess : TProcess;
  Buffer       : string;
  BytesRead    : integer;
  CurrentPart : integer = 0;
  SourcePart,DestPart : string;
  UmountOK : boolean = false;
  PreviousBlocks : int64 = 0;
  PartNum, PartNum2 : string;
  OutSATA : boolean; //true if output device is SATA
  complete : string;
  Logfiles, LogText, LogText2 : TStringList; //used to modify the logfiles generated during the clone
  LogFname : string;     //filename for the logfile being modified;
begin
  //debug
  if debug then DebugOut(true,'CloneDiskToDisk');
  //end debug
  ConfirmForm.close;
  MainFm.PageCtrl.Enabled:= false; //disable while running backup
  StartTime:= Now; //start time for partclone backup
  ProgressFm:=TProgressFm.Create(MainFm);
  //Hide the add note button, not needed for clone
  ProgressFm.AddBtn.Enabled:= false;
  ProgressFm.AddBtn.Visible:= false;
  //Enable the view log button
  ProgressFm.LogBtn.Enabled:= false;
  ProgressFm.LogBtn.Visible:= true;
  //Dangerous to cancel a clone partway through so disable ability to cancel
  ProgressFm.CancelBtn.Visible:= false;;
  ProgressFm.CancelBtn.Enabled:= false;;
  ProgressFm.Show;
  //delete log files in the default folder, doesn't look like deletefile accepts wildcards
  LogText:= TStringList.create;
  FindAllFiles(LogText, DefaultFolder, '*', false);
  for i:= 0 to LogText.count - 1 do DeleteFile(LogText[i]);
  LogText.free;
  //totalblocks is in clonedrive how is it populated?
  //used in backup, not in utilities?
  //set .selected false for partitions that will not be cloned
  for i:= 1 to CloneDrive.totalparts do
      if (CloneDrive.parts[i].ptype = 'extended') or (CloneDrive.parts[i].fs = 'swap') then
         CloneDrive.parts[i].selected:= false else CloneDrive.parts[i].selected:= true;
  for i:= 1 to CloneDrive.totalparts do
      if CloneDrive.parts[i].selected then TotalBlocks+= CloneDrive.parts[i].pused;
  ProgressFm.OverallProgBar.Max:= TotalBlocks;
  //populate the progress listbox
  ProgressFm.PartListBox.Items.Add('Partition table - not started');
  for i:= 1 to CloneDrive.totalparts do
      ProgressFm.PartListBox.Items.Add('Part ' + IntToStr(i) + ' - not started');
  //get the destination, e.g. sdc
  for i:= 0 to MainFm.ClChkBox2.Count - 1 do
      if MainFm.ClChkBox2.Checked[i] then Destination:= MainFm.ClChkBox2.Items[i];
  Destination:= trim(copy(Destination,1,pos(' : ',Destination)));
  //now check if destination is a SATA drive or NVME/SD card
  OutSATA:= pos('sd',Destination) = 1;
  ProgressFm.PartListBox.Items[0]:= 'Partition table - started';
  application.processmessages;
  //generate partition table for source drive
  cmdstr:= 'pkexec sfdisk --dump /dev/' + CloneDrive.mountpt + ' > ' + DefaultFolder + 'clone.sfdisk';
  setlength(buffer,bufsize);
  termCmd(cmdstr,cmdProcess); //run sfdisk
  //output from the command is on stderr
  BytesRead:= cmdProcess.stderr.Read(buffer[1],length(buffer));
  if BytesRead > 0 then
     begin
       setlength(buffer,BytesRead);
       MessageDlg('Error saving partition table for ' + cloneDrive.mountpt + ' to clone.sfdisk'
                  + LF + buffer,mtError,[mbOK],0);
       if settings.parttable then ProgressFm.PartListBox.Items[0]:= 'Partition table - error';
       exit;
     end;
  cmdProcess.Free;
  //modify the partition table for the new drive
  if not ModifyPTable(DefaultFolder + 'clone.sfdisk',DefaultFolder + 'clone.sfdisk',Destination)
  then exit;
  //restore grub
  GrubRestore(CloneDrive.mountpt,Destination,false);
  //now restore part table
  if not PTableRestore(DefaultFolder + 'clone.sfdisk',Destination) then exit;
  ProgressFm.PartListBox.Items[0]:= 'Partition table - complete';
  application.processmessages;
  //now run partclone
  for i:= 1 to CloneDrive.totalparts do if CloneDrive.parts[i].selected then
     begin
       CurrentPart+= 1;
       PstartTime:= Now;
       //set current progress bar
       ProgressFm.Caption:= 'Cloning Drive to Drive Part ' + IntToStr(CurrentPart) + ' of '
                            + IntToStr(CloneDrive.totalparts);
       //set the caption in the listbox
       ProgressFm.PartListBox.Items[CurrentPart]:= 'Part '+ IntToStr(CurrentPart) + ' - started';
       //get the partition number to be cloned - last char of pnumber
       PartNum:= copy(CloneDrive.parts[i].pnumber,length(CloneDrive.parts[i].pnumber),1);
       //just in case - check if more than 9 partitions
       PartNum2:= copy(CloneDrive.parts[i].pnumber,length(CloneDrive.parts[i].pnumber)-1,1);
       if pos(PartNum2,numbers) > 0 then PartNum:= PartNum2 + PartNum;
       //add the part number to destination, but if destination is NVME/SD then need to add in 'p'
       if OutSATA then DestPart:= Destination + PartNum else Destpart:= Destination + 'p' + PartNum;
       //also need to check going the other way, i.e. from nvme to sata, if so strip out the p
       SourcePart:= '/dev/';
       if pos('p',CloneDrive.parts[i].pnumber) = 1 then SourcePart+= copy(CloneDrive.parts[i].pnumber,2,99)
       else SourcePart+= CloneDrive.parts[i].pnumber;
       DestPart:= '/dev/'+ Destpart;
       cmdstr:= PartCloneC(SourcePart,DestPart,CloneDrive.parts[i].fs);
       //check if source part is mounted
       UmountOK:= CheckMount(CloneDrive.parts[i].pnumber) = '<none>';
       //need to umount the source partition if mounted
       if not UmountOK then if umount(CloneDrive.parts[i].pnumber) = 0 then UmountOK:= true;
       if UmountOK then //partition umounted, safe to proceed
          begin
            //run partclone and update progress form
            RunPartClone(cmdstr,CloneDrive.parts[i].pused,PreviousBlocks,TotalBlocks, StartTime);
            if CancelPartClone then
               begin
                 CancelPartClone:= false;
                 ProgressFm.Free;
                 exit;
               end;
            PreviousBlocks+= CloneDrive.parts[i].pused;
            if ProgressFm.CurrentProgLbl.Caption = 'Cloned successfully' then complete:= ' - complete '
            else complete:= ' - finished ';
            ProgressFm.PartListBox.Items[CurrentPart]:= 'Part '+ IntToStr(CurrentPart) + complete
                                                       + FormatDateTime('hh:nn:ss',Now - PstartTime);
            //modify the log file to add date and time
            Logtext:= TStringList.create;
            LogFname:= DefaultFolder + FormatDateTime('yyyymmdd',Now)
                       + '.' + replacestr(Destpart,'/dev/','') + '.log';
            Logtext.loadfromfile(LogFname); //looks like this overwrites the previous contents
            Logtext.insert(0,(copy(LogFname,rpos('/',LogFname)+1,99)));
            Logtext.insert(0,(FormatDateTime('dd mmm yyyy, hh:nn',Now)));
            Logtext.Add(separator);
            Logtext.savetofile(LogFname);
            Logtext.free;
          end
       else //not safe to restore partition, unable to umount it
          begin
            ProgressFm.PartListBox.Items[CurrentPart]:= 'Part '+ IntToStr(CurrentPart)
                                                        + ' - not restored';
            MessageDlg('Failed un-mounting ' + CloneDrive.parts[i].pnumber + LF +
                       'Not safe to clone a live partition, skipping...',mtError,[mbOK],0)
          end;
     end
  else //extended or swap partition
    begin
      CurrentPart+= 1;
      ProgressFm.Caption:= 'Cloning Drive to Drive Part ' + IntToStr(CurrentPart) + ' of '
                           + IntToStr(CloneDrive.totalparts);
      if CloneDrive.parts[i].ptype = 'extended' then
         ProgressFm.PartListBox.Items[CurrentPart]:= 'Part ' + IntToStr(CurrentPart)
                                                     + ' - skipped (extended)'
      else
        begin //create swap partition
          ProgressFm.PartListBox.Items[CurrentPart]:= 'Part '+ IntToStr(CurrentPart)+ ' - creating swap';
          PartNum:= copy(CloneDrive.parts[i].pnumber,length(CloneDrive.parts[i].pnumber),1);
          if OutSATA then DestPart:= Destination + PartNum else Destpart:= Destination + 'p' + PartNum;
          RestoreSwap(DestPart,getUUID(CloneDrive.parts[i].pnumber));
          ProgressFm.PartListBox.Items[CurrentPart]:= 'Part '+ IntToStr(CurrentPart)+ ' - swap created';
        end;
    end;
  //force the kernel to reread the partition table - just in case
  cmdstr:= 'pkexec partprobe /dev/' + Destination;
  termCmd(cmdstr,cmdProcess);
  setlength(buffer,bufsize);
  BytesRead:= cmdProcess.Output.Read(buffer[1],length(buffer));
  cmdProcess.free;
  //collect the logfiles together for viewing via the view log button
  LogText:= TStringList.create;
  //load the partition table
  LogText.loadfromfile(DefaultFolder + 'clone.sfdisk');
  LogText.insert(0,'PARTITION TABLE');
  LogText.insert(0,separator);
  LogText.insert(0,'Clone Drive to Drive ' + FormatDateTime('dd mmm yyyy, hh:nn',Now));
  LogText.add(separator);
  LogFiles:= TStringList.create;
  FindAllFiles(LogFiles, DefaultFolder, '*.log', false);
  LogFiles.sort;
  //load all the log files into a verify file
  if LogFiles.count >= 1 then
     for i:= 0 to LogFiles.count - 1 do
         begin
           LogText2:= TStringList.create;
           LogText2.loadfromfile(LogFiles[i]);
           for j:= 0 to LogText2.count - 1 do LogText.add(LogText2[j]);
           LogText2.free;
         end;
  LogText.savetofile(DefaultFolder + FormatDateTime('yyyymmdd',Now) + '.clone.log');
  LogText.free;
  LogFiles.free;

  //reset the progress form to indicate completion
  ProgressFm.Caption:= 'Clone complete';
  ProgressFm.PartcloneLbl.Caption:= '';
  ProgressFm.CancelBtn.Enabled:= false;
  ProgressFm.FinishBtn.Enabled:= true;
  ProgressFm.LogBtn.Enabled:= true;
end; //end procedure CloneDiskToDisk
{-------------------------------------------------------------------------------
PROCEDURE CLONEFILEOPEN
Opens the backup file containing the backup to be cloned to a new drive. Gets
the information in the backup file and populates listbox3 with the partitions in
the backup. Then populates checkbox4 with a list of the target drives.

V1.1 - an * added to the start of any partition in the checklistbox that contains an operating system.
For any swap partitions, strip out the UUID from the item in the checklistbox.

Parameters:
  Source - item index for the partition selected in the clone checkbox2.

Used by: Unit mainform - CloneBox2CK

V1.0 - 15 Oct 19
V1.1 - 8 Nov 19
V1.2 - 18 Jan 20, debug
V1.3 - 24 Sep 20, changes consequent on introducing the TRestPartData record. Uses builtpartstr to build
                  a text string to display in part check/list boxes
V1.4 - 03 Oct 20, debug code changed, bug - was using DefaultFolder as the initial folder for the open
                  dialog, should be MountFolder. chopstr on clone box3 title.
-------------------------------------------------------------------------------}
procedure CloneFileOpen(Source : integer);
var
  OpenDialog : TOpenDialog;
  ForceMount : boolean = false;
  MountOK : boolean = false;
  MountPoint : string;
  i : integer;
  maxlength : integer = 0;
begin
  //debug
  if debug then
     DebugOut(true,'CloneFileOpen','Source',IntToStr(Source),'Part clicked',BackDrive.parts[Source+1].pnumber);
  //end debug
  //check that an item or checkbox has been clicked
  with MainFm do
     begin
       //Cannot have backup files on an unknown or extended or swap partition
       with Backdrive do if (parts[Source+1].fs = '<unknown>')
          or (parts[Source+1].fs = 'swap')
          or (parts[Source+1].fs = 'extended') then exit;
       ClChkBox2.Checked[Source]:= true;
       MountPoint:= CheckMount(BackDrive.parts[Source+1].pnumber);
       if MountPoint = '<none>' then
          begin
            if mkdir(BackDrive.parts[Source+1].pnumber) = 0 then
               begin
                 if mount(BackDrive.parts[Source+1].pnumber,true) = 0 then
                    begin
                      MountPoint:= MountFolder + BackDrive.parts[Source+1].pnumber;
                      MountOK:= true;
                      ForceMount:= true; //should mean I unmount it when finished, but where
                    end
               end
          end
       else MountOK:= true;
       if MountOK then
          begin
            OpenDialog:= TOpenDialog.Create(PageCtrl.Page[2]);
            OpenDialog.Title:= 'Open backup file';
            OpenDialog.InitialDir:= MountPoint;
            OpenDialog.Filter := 'Backup files (*.backup)|*.backup';
            if OpenDialog.Execute then
               begin
                 RestFileName:= OpenDialog.Filename;
                 OpenDialog.Free;
                 getRestoreData(RestFileName,RSettings);
                 //irrespective of the settings in the backup file, for clone
                 //must restore the partition table
                 RSettings.PTable:= true;
                 ClListbox3Lbl.Caption:= '3 - Contents of ' + ChopStr(RestFileName,40);
                 ClListBox3.Enabled:= true;
                 CLListBox3.Clear; //in case it has already been populated
                 ClDateLbl.Visible:= true;
                 ClDateLbl.Caption:= 'Backup Date:  ' + RSettings.BackDate + LF
                                     + 'Time Now:       ' + FormatDateTime(FDateTimeStr,Now);
                 for i:= 1 to RSettings.TotalParts do
                     begin
                       ClListBox3.Items.Add(BuildPartStr(RSettings.NewParts[i]));
                       //add an * if OS in partition, [i-1] cos listbox indexed from zero
                       if length(ClListBox3.Items[i-1]) > maxlength
                       then maxlength:= length(ClListBox3.Items[i-1]);
                       //strip out the UUID from any item with a swap partition
                       if pos(': swap :',ClListBox3.Items[i-1]) > 0 then with ClListBox3 do
                          Items[i-1]:= copy(Items[i-1],1,pos(': swap :',Items[i-1]) + 6);
                     end;
                 //set horizontal scroll bar if needed
                 if maxlength * 6 > ChkBoxWidth then ClListBox3.ScrollWidth:= ChkBoxWidth + 1;
                 //if notes in backup make note button visible
                 if RSettings.Note then
                    begin
                      CloneNoteBtn.Enabled:= true;
                      CloneNoteBtn.Visible:= true;
                    end;
                 //populate the clone checklistbox4
                 //set all drives visible and selected drive not visible
                 for i:= 1 to TotalDrives do MyDrives[i].show:= true;
                 for i:= 0 to ClChkBox1.Count - 1 do if ClChkBox1.checked[i] then Source:= i;
                 MyDrives[Source + 1].show:= false;
                 //show drives in checkbox 4 and enable it
                 ClChkBox4.Enabled:= true;
                 ShowDrives(ClChkBox4);
               end;
          end;
     end;
end; //end procedure CloneFileOpen
{-------------------------------------------------------------------------------
PROCEDURE CLONEFILETODISK

Clones to a new drive from an existing backup of the 'source' drive by:

copying grub to the new drive - only required on mbr drives but done for all drive types.

modifying the part table file (*.sfdisk) for the new drive (on a GPT drive the backup partition table
sits in the last 34 sectors on the drive, lastlba needs modifying for the new drive). It creates a new
temporary part table file.

restoring the partition table from the temporary file.

using partclone to copy the partitions.

Used by: Unit confform - OkBtnCK

V1.1 - added functionality to create swap partition on target drive and skip any extended partitions
found.
v1.2 try moving grub and part restor around. no it is okay, grub first then part table
v1.3 changed because on stick not updating kernel on part change, so raise flag and msg at end
v1.4 problem found with systemd automatically using swap parts on boot, set swapoff in rc.local and
revert to previous version
v1.5 modified to clone to an NVME drive or SD card, need to add 'p' into the partition reference
V1.6 previously was using the loop counter, i, to determine the target partition #. Dangerous, part #s
may not be logical or sequential, e.g. sda2 = extended part, next part is sda5. Now extracts the source
part from RSettings.parts[i] and replaces source drive with target drive to get the target partition #.
V1.8 changes consequent on introducing TRestPartData record in place of a simple text str of partition
info.
V1.9 testing starttime:= time was returning 31 dec 99, whereas now returned the correct datetime, changed
     all occurences of time changed to now. Bug - was not checking to see if the destination was not
     a sata drive, fixed with outSATA.
     modified to handle nvme, a bug where it got the partitions on a destination nvme drive wrong

V1.0 - 15 Oct 2019
V1.1 - 30 Oct 2019
V1.3 - 01 Dec 2019
V1.4 - 03 Dec 2019
V1.5 - 18 Jan 2200, debug
V1.6 - 06 Mar 2020
V1.7 - 22 Mar 2020
V1.8 - 24 Sep 2020
V1.9 - 23 Jan 2021  .
V1.10 - 19 Feb 2021, another fix missing the p when sata to nvme/sd card on the logfile name
-------------------------------------------------------------------------------}
procedure CloneFileToDisk;
const
  separator = '-----------------------------------------------------------------------------';
var
  cmdstr : string; //restore command line
  StartTime, PstartTime : TDateTime; //start time for clone & start time for each partition
  TotalBlocks : int64 = 0; //total blocks to restore
  PreviousBlocks : int64 = 0; //blocks already restored
  i, j : integer;
  TargetNum : integer = 0;
  Target, complete : string; //target drive in the form sda, sdb, etc.
  UmountOK : boolean;
  LogText : TStringList;
  LogText2 : TStringList;
  LogFname : string;
  Destination : string;
  DestPartnum : string; //e.g. the '1' bit of 'sda1'
  LogFiles : TStringList;
  OutSATA : boolean; //true if output device is SATA
begin
  //debug
  if debug then DebugOut(true,'CloneFileToDisk');
  //end debug
  StartTime:= Now;
  ConfirmForm.Close;
  //disable the main form
  MainFm.PageCtrl.Enabled:= false;
  ProgressFm:=TProgressFm.Create(MainFm);
  //Hide the add note button, not needed for restore
  ProgressFm.AddBtn.Enabled:= false;
  ProgressFm.AddBtn.Visible:= false;
  //Enable the view log button
  ProgressFm.LogBtn.Enabled:= false;
  ProgressFm.LogBtn.Visible:= true;
  //delete log files in the default folder, doesn't look like deletefile accepts wildcards
  LogText:= TStringList.create;
  FindAllFiles(LogText, DefaultFolder, '*', false);
  for i:= 0 to LogText.count - 1 do DeleteFile(LogText[i]);
  LogText.free;
  //Dangerous to cancel a restore partway through so disable ability to cancel
  ProgressFm.CancelBtn.Enabled:= false;;
  ProgressFm.CancelBtn.Visible:= false;;
  ProgressFm.Show;
  ProgressFm.Caption:= 'Restoring partition table';
  ProgressFm.PartListBox.Items.Add('Partition table - not started');
  for i:= 1 to RSettings.TotalParts do
      begin
        TotalBlocks+= RSettings.NewParts[i].blocks;
        ProgressFm.PartListBox.Items.Add('Part ' + IntToStr(i) + ' - not started');
      end;
  ProgressFm.OverallProgBar.Max:= TotalBlocks;
  //get the target drive
  for i:= 0 to MainFm.ClChkBox4.Items.Count - 1 do if MainFm.ClChkBox4.Checked[i]
  then TargetNum:= i;
  Target:= MainFm.ClChkBox4.Items[TargetNum]; //returns a drive description
  Target:= trim(copy(Target,1,pos(':',Target)-1)); //strip out the mount point, e.g. sda
  RSettings.DestMountpt:= Target;
  //now check if destination is a SATA drive or NVME/SD card
  OutSATA:= pos('sd',RSettings.DestMountpt) = 1;
  //restore the partition table and grub
  ProgressFm.PartListBox.Items[0]:= 'Partition table - started';
  application.processmessages;
  //modify the partition table for the new drive
  if not ModifyPTable(ReplaceStr(RestFileName,'.backup','.sfdisk'),DefaultFolder + 'clone.sfdisk',Target)
     then exit;
  //restore grub
  GrubRestore(ReplaceStr(RestFileName,'.backup','.grub'),Target,true);
  //restore partition table
  if not PTableRestore(DefaultFolder + 'clone.sfdisk',Target) then exit;
  ProgressFm.PartListBox.Items[0]:= 'Partition table - complete ';
  application.processmessages;
  //check if SATA or NVME/SD card, if NVME add 'p' to target
  if pos('sd',Target) <> 1 then Target:= Target + 'p';
  //now run partclone
  for i:= 1 to RSettings.TotalParts do
     begin
       if RSettings.NewParts[i].blocks <> 0 then //clone the partition
          begin
            PstartTime:= Now;
            //set current progress bar
            ProgressFm.Caption:= 'Cloning File to Drive Part ' + IntToStr(i) + ' of '
                                 + IntToStr(RSettings.TotalParts);
            //set the caption in the listbox
            ProgressFm.PartListBox.Items[i]:= 'Part '+ IntToStr(i) + ' - started';
            //got errors running partclone when the system had automounted the target
            //partition even after wiping and resetting the part table, so check and unmount

            //note already had destination as variable -check what used for

            //this is okay for nvme but if partitions are not numbered concurrently it has a problem
            //have to work with pnumber  and Mountpt
            DestPartnum:= replacestr(RSettings.NewParts[i].pnumber,RSettings.Mountpt,'');
            UmountOK:= CheckMount(Target + DestPartnum) = '<none>';
            if not UmountOK then UmountOK:= Umount(Target + DestPartnum) = 0;
            if UmountOK then
               begin
                 //build the partclone command line
                 if OutSATA then Destination:= RSettings.NewParts[i].pnumber else
                   Destination:= RSettings.NewParts[i].pnumber;
                 cmdstr:= PartCloneR(RestFilename,RSettings.NewParts[i].pnumber,RSettings.NewParts[i].fs);
                 //run partclone and update progress form
                 RunPartClone(cmdstr,RSettings.NewParts[i].blocks,PreviousBlocks,TotalBlocks, StartTime);
                 PreviousBlocks+= RSettings.NewParts[i].blocks;
                 if ProgressFm.CurrentProgLbl.Caption = 'Cloned successfully' then
                    complete:= ' - complete ' else complete:= ' - finished ';
                 ProgressFm.PartListBox.Items[i]:= 'Part '+ IntToStr(i) + complete
                                                   + FormatDateTime('hh:nn:ss',Now - PstartTime);
                 //modify the log file to add date and time
                 Logtext:= TStringList.create;
                 //  PartNum:= ReplaceStr(PartNum,RSettings.Mountpt,RSettings.DestMountpt);
                 with RSettings do Destination:= ReplaceStr(NewParts[i].pnumber,Mountpt,DestMountpt);
                 // check if cloning from sd/nvme to sata
                 if (pos('sd',RSettings.Mountpt) =0) and (pos('sd',RSettings.DestMountpt) = 1) then
                    Destination:= replacestr(Destination,RSettings.DestMountpt + 'p',RSettings.DestMountpt);
                 // check if cloning from sata to sd/nvme - need to add a 'p'
                 if (pos('sd',RSettings.Mountpt) = 1) and (pos('sd',RSettings.DestMountpt) = 0) then
                    begin
                      writeln('clone from file, destination=',Destination);
                      Destination:= replacestr(Destination,RSettings.DestMountpt,RSettings.DestMountpt+'p');
                    end;


                 LogFname:= DefaultFolder + FormatDateTime('yyyymmdd',Now)
                            + '.' + Destination + '.log';
                 writeln('clone from file, logfilename=',LogFname);
                 Logtext.loadfromfile(LogFname);
                 Logtext.insert(0,(copy(LogFname,rpos('/',LogFname)+1,99)));
                 Logtext.insert(0,(FormatDateTime('dd mmm yyyy, hh:nn',Now)));
                 Logtext.Add(separator);
                 Logtext.savetofile(LogFname);
                 Logtext.free;
               end
          end
       else //either a swap partition or an extended partition
         begin
           if RSettings.NewParts[i].ptype = 'extended' then //it's an extended partition
              ProgressFm.PartListBox.Items[i]:= 'Part ' + IntToStr(i) + ' - extended (skipped)'
           else
             if RSettings.NewParts[i].fs = 'swap' then //recreate swap partition
                 begin
                  with RSettings do SwapPart:= ReplaceStr(SwapPart,Mountpt,Target);
                  RestoreSwap(RSettings.SwapPart,RSettings.SwapUUID);
                  ProgressFm.PartListBox.Items[i]:= 'Part ' + IntToStr(i) + ' - swap recreated';
                end
             else ProgressFm.PartListBox.Items[i]:= 'Part ' + IntToStr(i) + ' - not included';
         end;
     end;
  //collect the logfiles together for viewing via the view log button
  LogText:= TStringList.create;
  //load the partition table
  LogText.loadfromfile(DefaultFolder + 'clone.sfdisk');
  LogText.insert(0,'PARTITION TABLE');
  LogText.insert(0,separator);
  LogText.insert(0,'Clone File to Drive ' + FormatDateTime('dd mmm yyyy, hh:nn',Now));
  LogText.add(separator);
  LogFiles:= TStringList.create;
  FindAllFiles(LogFiles, DefaultFolder, '*.log', false);
  LogFiles.sort;
  //load all the log files into a verify file
  if LogFiles.count >= 1 then
     for i:= 0 to LogFiles.count - 1 do
         begin
           LogText2:= TStringList.create;
           LogText2.loadfromfile(LogFiles[i]);
           for j:= 0 to LogText2.count - 1 do LogText.add(LogText2[j]);
           LogText2.free;
         end;
  LogText.savetofile(DefaultFolder + FormatDateTime('yyyymmdd',Now) + '.clone.log');
  LogText.free;
  LogFiles.free;
  //enable the log button
  ProgressFm.LogBtn.Enabled:= true;
  //reset the progress form to indicate completion
  ProgressFm.Caption:= 'Clone from file complete';
  ProgressFm.PartcloneLbl.Caption:= '';
  ProgressFm.CancelBtn.Enabled:= false;
  ProgressFm.FinishBtn.Enabled:= true;
  //enable the main form
  //MainFm.PageCtrl.Enabled:= true; not needed done in the finish button on the prog form
end; //end procedure CloneFileToDisk
{-------------------------------------------------------------------------------
PROCEDURE CLONELOG

Displays the log files after a clone.

Used by: Unit progform - ProgFmViewLog

V1.0 - 18 Jan 2021
-------------------------------------------------------------------------------}
procedure CloneLog;
var
    CloneNote : TNoteForm;
    CloneFiles : TStringList;
begin
  //debug
  if debug then DebugOut(true,'CloneLog');
  //end debug
  //somewhere need to free noteform????
  //There should only be one *.clone.log file
  CloneFiles:= TStringList.create;
  FindAllFiles(CloneFiles, DefaultFolder, '*.clone.log', false);
  if CloneFiles.count = 1 then
     begin
       CloneNote:= TNoteForm.create(ProgressFm);
       with CloneNote do
            begin
              SaveButton.Enabled:= false;
              SaveButton.Visible:= false;
              CancelButton.Left:= 400;
              Caption:= 'Clone Log'; //caption for the title bar
              Width:= 500;
              BackMemo.Width:= 460;
              BackMemo.Font.Name:= 'Noto Sans Mono';
              BackMemo.Font.Size:= 9;
              BackMemo.WordWrap:= false;
              BackMemo.Lines.LoadFromFile(CloneFiles[0]);
              CloneFiles.free;
              show;
            end;
     end;
end;

end.

