{ FoxClone
  Copyright (C) 2019, 2020 Andy Hardwick
  This source is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.
  This code is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.
  A copy of the GNU General Public License is available on the World Wide Web
  at <http://www.gnu.org/copyleft/gpl.html>. You can also obtain it by writing
  to the Free Software Foundation, Inc., 51 Franklin Street - Fifth Floor,
  Boston, MA 02110-1335, USA.
}

{-------------------------------------------------------------------------------
UNIT MAINFORM
Mainform creates all the objects on the mainform for each of the four tabs -
backup, restore, clone and settings; plus their associated event handlers.

Changes:

V1.1 - 12 Jan 2020, changed to add 'about' tab and the progress form
V1.2 - 14 Feb 2020, procedures/functions in settings.inc merged in and directive to include
       settings.inc. removed. procedure SettingsPartTable deleted.
V1.3 - 6 June 2020, xxxxxx
V1.4 - 17 Sep, 2020, hints moved into constants.
V1.5 - 11 Jan, 2021, tabs moved into separate .inc files at the time the verify tab was created.

-------------------------------------------------------------------------------}
unit mainform;

{$mode objfpc}{$H+}

interface

uses
  Interfaces, Classes, SysUtils, Forms, Controls, Graphics, Dialogs, ComCtrls,
  CheckLst, StdCtrls, dateutils, strutils, process, utilities,
  ExtCtrls, backup, restore, noteform, confform, progform, clone, startupform,
  LCLIntf, verify;

type
  TMainFm = class(TForm)
    VersionLbl : TLabel;
    PageCtrl : TPageControl; //setup the tabs
    {---------------------------------------------------------------------------
    Graphics objects for images on the form
    ---------------------------------------------------------------------------}
    DriveImg, FileImg, ArrowImg, SettingsImg, CSourceImg, CDestImg, FoxImg, CheckImg : TImage;
    {---------------------------------------------------------------------------
    Objects on the backup tab
    ---------------------------------------------------------------------------}
    BackTitleLbl : Tlabel; //title for the top of the tab
    BackBox1Lbl,      //title for the source drive listbox = drive to backup
    BackBox2Lbl,       //title for the source partition listbox = parts to backup
    BackBox3Lbl,      //title for the destination drive listbox
    BackBox4Lbl : TLabel; //title for the destination partition listbox
    BackChkBox1,             //To list source drives
    BackChkBox2,              //To list source partitions on selected drive
    BackChkBox3,               //To list destination drive
    BackChkBox4: TCheckListBox; //To list destination partition on destination drive
    NextBtn,   //button to open the file dialogue box for the folder & filename
    QuitBtn : TButton;   //Quit button to terminate, used on all tabs
    {---------------------------------------------------------------------------
    Objects on the restore tab
    ---------------------------------------------------------------------------}
    RestTitleLbl : TLabel; //title for the top of the tab
    RestBox1Lbl : TLabel;
    RestChkBox1 : TCheckListBox;
    RestBox2Lbl : TLabel;
    RestChkBox2 : TCheckListBox;
    RestBox3Lbl : TLabel;
    RestChkBox3 : TCheckListBox;
    RDateLbl : TLabel; //date the backup was taken & current time
    RestBox4Lbl : TLabel;
    RestLstBox4 : TListBox;
    NoteBtn : TButton;
    RestBtn : TButton;
    RDestInfoLbl : TLabel;
    {---------------------------------------------------------------------------
    Objects on the clone tab
    ---------------------------------------------------------------------------}
    CloneLbl : TLabel; //title for the top of the tab
    CloneComboBox : TComboBox;
    ClCheckBox1Lbl : TLabel;
    ClChkBox1 : TCheckListBox;
    ClCheckbox2Lbl : TLabel;
    ClChkbox2 : TCheckListBox;
    ClListbox3Lbl : TLabel;
    ClListbox3 : TListBox;
    ClDateLbl : TLabel;
    CloneBtn : TButton;
    CloneNoteBtn : TButton;
    ClCheckBox4Lbl : TLabel;
    ClChkBox4 : TCheckListBox;
    {---------------------------------------------------------------------------
    Objects on the verify tab
    ---------------------------------------------------------------------------}
    VerifyLbl : TLabel; //title for the top of the tab
    VBox1Lbl : TLabel; //title for checkbox1
    VChkBox1 : TCheckListBox; //drive select checklistbox
    VBox2Lbl : TLabel; //title for checkbox 2
    VChkBox2 : TCheckListBox; //partition select checklistbox
    VBox3Lbl : TLabel; //title for checkbox 3
    VChkBox3 : TCheckListBox; //checklist box containing the content of the backup file
    VDateLbl : TLabel; //label containing backup date and time
    VNoteBtn : TButton; //button to view notes on the backup
    VerifyBtn : TButton; //button to start the verify process
    {---------------------------------------------------------------------------
    Objects on the settings tab
    ---------------------------------------------------------------------------}
    SettingsLbl : TLabel;
    PartTableCheckBox : TCheckBox;
    CompressionChkBox : TCheckBox;
    CompLvlComboBox : TComboBox;
    SplitCheckBox : TCheckBox;
    SplitFileSize : TEdit;
    ResetBtn : TButton;
    SQuitBtn : TButton;
    {---------------------------------------------------------------------------
    Objects on the about tab
    ---------------------------------------------------------------------------}
    TitleLbl : TLabel;
    DescribeLbl : TLabel;
    CopyRLbl : TLabel;
    LicenceLbl : TLabel;

  const
      //default settings for the checkboxes on the backup, restore & clone tabs
      ChkBoxHeight = 150;
      ChkBoxWidth = 360;
      TopChkBox1 = 100;  //to top of the first row of checkboxes
      TopChkBox2 = 300;  //to top of the second row of checkboxes
      LeftChkBox1 = 15;  //left edge first column of checkboxes
      LeftChkBox2 = 410; //left edge second column of checkboxes
      //hints when hovering over objects
      PartitionHint = 'Mount point' + LF +
                      'File system' + LF +
                      '[label]' + LF +
                      'Size' + LF +
                      '% used' + LF +
                      'Flags - if set' + LF +
                      'Type - legacy only, or' + LF +
                      '(Partition name) - GPT only' + LF +
                      '*OS* - if one';
      DriveHint = 'Mount point' + LF +
                  'Model name' + LF +
                  'Drive size' + LF +
                  'Partition table type' + LF +
                  'Only a single drive can be selected';


    public
      constructor Create(TheOwner: TComponent); override;
      procedure ReadDrives(Sender: TObject); //as well as being used in the constructor, used after
                                             //restore and clone in progform.ProgFmFinish
    private
      //setup individual tabs
      procedure CreateBackupPage;
      procedure CreateRestorePage;
      procedure CreateClonePage;
      procedure CreateVerifyPage;
      procedure CreateSettingsPage;
      procedure CreateAboutPage;
      //backup
      procedure BackBox1CK(Sender: TObject);
      procedure BackBox2CK(Sender: TObject);
      procedure BackBox3CK(Sender: TObject);
      procedure BackBox4CK(Sender: TObject);
      procedure BackSaveCK(Sender: TObject);
      procedure QuitCK(Sender: TObject);
      //restore
      procedure RestBox1CK(Sender: TObject);
      procedure RestBox2CK(Sender: TObject);
      procedure RestBox3CK(Sender: TObject);
      procedure RestNoteCK(Sender: TObject);
      procedure RestRestoreCK(Sender: TObject);
      //clone
      procedure CloneType(Sender: TObject);
      procedure CloneBox1CK(Sender: TObject);
      procedure CloneBox2CK(Sender: TObject);
      procedure CloneBtnCK(Sender: TObject);
      procedure CloneBox4CK(Sender: TObject);
      //verify
      procedure VBox1CK(Sender: TObject);
      procedure VBox2CK(Sender: TObject);
      procedure VNoteCK(Sender: TObject);
      procedure VBox3CK(Sender: TObject);
      procedure VButtonCK(Sender: TObject);

      //settings
      procedure SettingsCompression(Sender: TObject);
      procedure SettingsCompLevel(Sender: TObject);
      procedure SettingsSplit(Sender: TObject);
      procedure SettingsSplitSize(Sender: TObject);
      procedure SettingsDefault(Sender: TObject);
  end;

var
  MainFm: TMainFm;
  FormActivated : boolean; //as well as being used in the constructor, used after
                           //restore and clone in progform.ProgFmFinish

implementation

{$R *.lfm}
{-------------------------------------------------------------------------------
CONSTRUCTOR
Creates the main window and the tabs for backup, restore, clone & settings
Calls procedures to set up each individual tab and start with the backup tab.

Also runs GetDriveInfo to read the system and get information on all drives and
partitions, and sets the initial defaults for compression, split files, etc.

V1.1 - VersionLbl changed to use function GetVersion to get the current version,
current kernel and version date.

V1.2 - porting onto the iso - ubuntu was using sans as the default font, so changes here
to make sure that foxclone is using noto sans as was originally designed and all the forms look the
same as designed and displayed on LM19 cinnamon. mainform caption changed to just foxclone

V1.0 - 27 May 19
V1.1 - 6 Nov 19
V1.2 - 22 Nov 19
V1.3 - 30 Dec 19 - debug code
V1.4 - 30 Jan 20 - about tab
V1.5 - 24 Sep 2020 - BorderStyle set to stop window resizing
-------------------------------------------------------------------------------}
constructor TMainFm.Create(TheOwner: TComponent);
const
  wHeight = 600;  //main window initial height
  wWidth = 800;   //main window initial width
var
  i : integer;
begin
  //debug
  if debug then DebugOut(true,'constructor TMainFm.Create');
  //end debug
  inherited CreateNew(TheOwner,1);
  //font properties added to ensure that windows look the same on other platforms,
  //eg. iso based on ubuntu
  Font.Name:= 'Noto Sans';
  Font.Size:= 10;
  InitialSettings; //for backing up parttable, compressing files, etc.
  //setup date/time format
  if FormatSettings.ShortDateFormat = 'dd/mm/yy' then FDateTimeStr:= 'dd mmm yyyy, hh:nn'
  else FDateTimeStr:= 'mmm dd yyyy, hh:nn';
  Caption:= 'Foxclone'; //caption for the title bar
  Position:=poScreenCenter;
  SetInitialBounds(0,0,wWidth,wHeight); //defines initial size
  BorderStyle:= bsSingle;
  OnActivate:= @ReadDrives; //displays popup progress window and reads all drives in the system
  //setup the tabs in the main window
  PageCtrl:=TPageControl.Create(MainFm);
  with PageCtrl do
    begin
      //Name:='PageCtrl';
      Parent:=Self;
      Align:=alTop; //may need changing if we put in a menu bar
      //leave some space at the bottom of the tab windows for a status bar
      Height:=Parent.ClientHeight-40;
      for i:=1 to 6 do
          with TTabSheet.Create(PageCtrl) do
               begin
                 PageControl:=PageCtrl;
                 Caption:=Name;
                 case i of
                      1 : Name:= 'Backup';
                      2 : Name:= 'Restore';
                      3 : Name:= 'Clone';
                      4 : Name:= 'Verify';
                      5 : Name:= 'Settings';
                      6 : Name:= 'About';
                 end;
               end;
      Anchors:=[akLeft,akTop,akRight,akBottom];
    end;
  VersionLbl := TLabel.Create(Self);
  with VersionLbl do
       begin
         Parent:= MainFm;
         Top:= 570;
         Left:= 580;
         Width:= 200;
         Height:= 16;
         Font.Size:= 9; //added after problems with display on the iso - was falling off the edge
         Alignment:= taRightJustify; //doesn't work
         Caption:= GetVersion;
         Show;
       end;
  CreateBackupPage;
  CreateRestorePage;
  CreateClonePage;
  CreateVerifyPage;
  CreateSettingsPage;
  CreateAboutPage;
  PageCtrl.PageIndex:=0; //this controls which tab is selected on start

end; //end constructor

{-------------------------------------------------------------------------------
PROCEDURE READDRIVES
Event handler for the onshow event for the main form. Runs procedures to read the drives with parted,
get the block sizes for all mountable partitions and labels, runs osprober to find any installed
operating systems on the system.

Uses startupform to display a progress window while getting drive info.

Information is stored the the MyDrives array (see unit utilities) which is then used by the
rest of the program.

Used by: the mainform constructor Create

V1.0 - 12 Jan 2020
V1.1 - 28 Feb 2020 - modified to create the mountfolder on startup
V1.2 - 20 Mar 2020 - finds the root partition and boot drive so that BootDrive can be later checked on
                     backup (as a source) or clone (as a destination) and a warning displayed.
V1.3 - 18 Jan 2021 - main form disabled on start, enabled on finish.
                     creates the DefaultFolder if needed.
                     changed use of mkdir on MountFolder to CreateDir
-------------------------------------------------------------------------------}
procedure TMainFm.ReadDrives(Sender: TObject);
var
  finish : TDateTime;
  debugtxt : string;
  i,j : integer;
begin
  //debug
  if debug then
     begin
       if FormActivated then debugtxt:= 'true' else debugtxt:= 'false';
       DebugOut(true,'TMainFm.ReadDrives','Activated',debugtxt);
     end;
  //end debug
  if not FormActivated then
     begin
       FormActivated:= true; //so this is only run once
       MainFm.enabled:= false;
       StartupFm:= TStartupFm.create(MainFm);
       StartupFm.Show;
       //create mount folder, cannot use the mkdir function in utilities - it automatically creates
       //the folder in MountFolder
       if not DirectoryExists(MountFolder) then CreateDir(MountFolder);
       //create temp folder, likewise cannot use mkdir.
       if not DirectoryExists(DefaultFolder) then CreateDir(DefaultFolder);
       //get system info
       StartupFm.ProgLbl.Caption:= 'Running parted...';
       Application.ProcessMessages;
       parted(MyDrives,TotalDrives);
       getSerials(MyDrives,TotalDrives); //added in foxclone V40, gets serial #s on drives.

       StartupFm.ProgBar.Position:= 25;
       StartupFm.PartLb1.Caption:= StartupFm.PartLb1.Caption + ' ✔';

       StartupFm.ProgLbl.Caption:= 'Getting partition sizes...';
       Application.ProcessMessages;
       GetPartSize(MyDrives,TotalDrives);
       StartupFm.ProgBar.Position:= 50;
       StartupFm.SizeLbl.Caption:= StartupFm.SizeLbl.Caption + ' ✔';

       StartupFm.ProgLbl.Caption:= 'Getting partition labels...';
       Application.ProcessMessages;
       partLabels(MyDrives,TotalDrives);
       StartupFm.ProgBar.Position:= 75;
       StartupFm.LabelLbl.Caption:= StartupFm.LabelLbl.Caption + ' ✔';
       StartupFm.ProgLbl.Caption:= 'Searching for operating systems...';
       Application.ProcessMessages;
       OSprobe;
       StartupFm.OSLbl.Caption:= StartupFm.OSLbl.Caption + ' ✔';
       StartupFm.ProgBar.Position:= 100;
       StartupFm.ProgLbl.Caption:= 'Done';
       Application.ProcessMessages;

       //populate the first checkboxes with drive info on backup, restore and clone tabs
       ShowDrives(MainFm.BackChkBox1);
       ShowDrives(MainFm.RestChkBox1);
       ShowDrives(MainFm.ClChkBox1);
       ShowDrives(MainFm.VChkBox1);

       //find the boot drive
       for i:= 1 to TotalDrives do
           for j:= 1 to MyDrives[i].totalparts do
               if CheckMount(MyDrives[i].parts[j].pnumber) = '/' then BootDrive:= MyDrives[i].mountpt;
       finish:= now;
       //put a small delay in before the form closes
       repeat
       until SecondsBetween(now,finish) > 1;
       StartupFm.Close;
       MainFm.enabled:= true;
     end;
end; //end procedure ReadDrives

{-------------------------------------------------------------------------------
Include directives to reduce the size of the mainform unit and make it more manageable.
Each tab on mainform (except for the about page) has its own inc file.
-------------------------------------------------------------------------------}

{$I backtab.inc}
{$I resttab.inc}
{$I clonetab.inc}
{$I verifytab.inc}
{$I settingstab.inc}

{-------------------------------------------------------------------------------
PROCEDURE CREATEABOUTPAGE
Creates the about page objects

Used by: the mainform constructor Create

V1.0 - 30 Jan 20
-------------------------------------------------------------------------------}
procedure TMainFm.CreateAboutPage;
const
  marginleft = 15;
var
  kernel : string;
begin
  kernel:= getVersion;
  kernel:= copy(kernel,pos('k',kernel)+1,99);
  kernel:= copy(kernel,1,pos(',',kernel)-1);
  {---------------------------------------------------------------------------
  Image and main title
  ---------------------------------------------------------------------------}
  FoxImg:=TImage.Create(Self);
  with FoxImg do begin
       Parent:=PageCtrl.Page[5];
       Left:= marginleft;
       Width:= 96;
       Height:= 96;
       Top:= 30;
       Picture.LoadFromFile(FoxImgFile);
  end;
  //Main title for the tab
  TitleLbl:= TLabel.Create(Self);
  with TitleLbl do
       begin
         Parent:= PageCtrl.Page[5];
         Top:= 30;
         Left:= 120;
         Width:= 150;
         Height:= 16;
         Caption:= 'Foxclone';
         Font.Size:= 14;
       end;
  {---------------------------------------------------------------------------
  Description
  ---------------------------------------------------------------------------}
  DescribeLbl:= TLabel.Create(Self);
  with DescribeLbl do
       begin
         Parent:= PageCtrl.Page[5];
         Top:= TitleLbl.Top + 30;
         Left:= 120;
         Width:= 150;
         Height:= 16;
         Caption:= 'An image backup, restore && clone tool. Foxclone is a front end for partclone ' + LF+
                   'written in Pascal using the lazarus IDE.' + LF +
                   'The ISO is a minimal custom ubuntu running a ' + kernel + ' kernel.';
         Font.Size:= 10;
       end;
  {---------------------------------------------------------------------------
  Copyright label
  ---------------------------------------------------------------------------}
  CopyRLbl:= TLabel.Create(Self);
  with CopyRLbl do
       begin
         Parent:= PageCtrl.Page[5];
         Top:= DescribeLbl.Top + 70;
         Left:= 120;
         Width:= 150;
         Height:= 16;
         Caption:= 'Copyright © 2019—2021: Andy Hardwick, Larry Hale.' + LF +
                   'Images ShiipArt && gaga.vastard.';
         Font.Size:= 10;
       end;
  {---------------------------------------------------------------------------
  Licence
  ---------------------------------------------------------------------------}
  LicenceLbl:= TLabel.Create(Self);
  with LicenceLbl do
       begin
         Parent:= PageCtrl.Page[5];
         Top:= CopyRLbl.Top + 55;
         Left:= 120;
         Width:= 150;
         Height:= 16;
         Caption:= 'This program is free software: you can redistribute it and/or modify' + LF +
        'it under the terms of the GNU General Public License as published by' + LF +
        'the Free Software Foundation, either version 3 of the License, or'   + LF +
        '(at your option) any later version.'     + LF + LF +

        'This program is distributed in the hope that it will be useful,'  + LF +
        'but WITHOUT ANY WARRANTY; without even the implied warranty of'  + LF +
        'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the'  + LF +
        'GNU General Public License for more details http://www.gnu.org/licenses/.' ;
         Font.Size:= 10;
       end;
end; //procedure CreateAboutPage


end.

