{ FoxClone
  Copyright (C) 2019 - 2021 Andy Hardwick
  This source is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.
  This code is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.
  A copy of the GNU General Public License is available on the World Wide Web
  at <http://www.gnu.org/copyleft/gpl.html>. You can also obtain it by writing
  to the Free Software Foundation, Inc., 51 Franklin Street - Fifth Floor,
  Boston, MA 02110-1335, USA.
}

{-------------------------------------------------------------------------------
UNIT VERFIFY
Contains verify specific procedures.

-------------------------------------------------------------------------------}
unit verify;

{$mode objfpc}{$H+}

interface

procedure VBox1(Source : integer);
procedure VBox2(Source : integer);
procedure VGetFile(Source : integer);
procedure VerifyParts;
procedure VerifyLog;

implementation
uses
  Interfaces, Classes, SysUtils, Forms, Controls, Graphics, Dialogs, ComCtrls,
  CheckLst, StdCtrls, dateutils, strutils, process, utilities, mainform, progform, fileutil, noteform;
{-------------------------------------------------------------------------------
PROCEDURE VBOX1
Resets boxes 2 - 4 on the verify page.

Parameters:
  Source - the item clicked in the source drive checkbox

Used by: Unit MainFm - VBox1CK

V1.0 - 11 Jan 2021
-------------------------------------------------------------------------------}
procedure VBox1(Source : integer);
var
  i : integer;
  maxlength : integer = 0;
  partstr : string;
begin
  //debug
  if debug then DebugOut(true,'VBox1','Source',IntToStr(Source));
  //end debug
  with MainFm do
       begin
         VChkBox1.CheckAll(cbUnchecked);
         VChkBox1.Checked[Source]:= true;
         VChkBox2.Enabled:= true;
         VChkBox2.Clear;
         //reset everything
         VChkBox3.Clear;
         VChkBox3.Enabled:= false;
         VChkBox3.ScrollWidth:= ChkBoxWidth;
         VBox3Lbl.Caption:= '3 - Contents of Backup:';
         VNoteBtn.Enabled:= false;
         VNoteBtn.Visible:= false;
         VDateLbl.Visible:= false;
         VerifyBtn.enabled:= false;
       end;
  BackDrive:= MyDrives[Source + 1]; //set the drive for the verify source
  //populate checkbox2 with the partitions on the source drive
  for i:= 1 to BackDrive.totalparts do
      begin
        partstr:= PartInfo(BackDrive.parts[i],BackDrive.ptabletype);
        MainFm.VChkBox2.Items.Add(partstr);
        if length(partstr) > maxlength then maxlength:= length(partstr);
      end;
  //set scroll bars on the source partition checklistbox
  with MainFm do if maxlength * 6 > ChkBoxWidth
                 then VChkBox2.ScrollWidth:= ChkBoxWidth + 1
                 else VChkBox2.ScrollWidth:= ChkBoxWidth;
end; //end procedure RestBox1
{-------------------------------------------------------------------------------
PROCEDURE VBOX2
Checks for a valid partition type that contains the backup files. If the partition
is valid, enables the checkbox.
Runs the procedure to open the backup file and get the backup data.

Parameters:
  Source - the item clicked in the source partition checkbox

Used by: Unit MainFm - VBox2CK

V1.0 - 12 Jan 2021
-------------------------------------------------------------------------------}
procedure VBox2(Source : integer);
begin
  //debug
  if debug then DebugOut(true,'VBox2','Source',IntToStr(Source));
  //end debug
  MainFm.VChkBox2.CheckAll(cbUnchecked);
  //Cannot have backup files on an unknown or extended or swap partition
  with Backdrive do if (parts[Source+1].fs = '<unknown>')
       or (parts[Source+1].fs = 'swap')
       or (parts[Source+1].ptype = 'extended') then exit;
  MainFm.VChkBox2.Checked[Source]:= true;
  VGetFile(Source);
end; //end procedure VBox2
{-------------------------------------------------------------------------------
PROCEDURE VGETFILE
Copied/modified version of restgetfile.
Checks that partition selected is mounted, if not mounts it.
Prompts for and opens the backup file.
Reads the contents of the backup file and populates the RSettings record with
the information in file.
Populates checkbox 3 with the partitions in the backup set (as defined by the
backup file).

Parameters:
  Source - the item clicked in the source partition checkbox

Used by: VBox2

V1.0 - 12 Jan 2021
-------------------------------------------------------------------------------}
procedure VGetFile(Source : integer);
var
  OpenDialog : TOpenDialog;
  ForceMount : boolean = false;
  MountOK : boolean = false;
  MountPoint : string;
  i : integer;
  maxlength : integer = 0;
begin
  //debug
  if debug then DebugOut(true,'VGetFile','Source',IntToStr(Source));
  //end debug
  MountPoint:= CheckMount(BackDrive.parts[Source+1].pnumber);
  if MountPoint <> '<none>' then MountOK:= true
  else
    begin
      if mkdir(BackDrive.parts[Source+1].pnumber) = 0 then
         begin
           if mount(BackDrive.parts[Source+1].pnumber,true) = 0 then
              begin
                MountPoint:= MountFolder + BackDrive.parts[Source+1].pnumber;
                MountOK:= true;
                ForceMount:= true;
              end
         end
    end;
  if not MountOK then exit;
  //everything okay so get the backup file
  OpenDialog:= TOpenDialog.Create(MainFm);
  OpenDialog.Title:= 'Open backup file';
  OpenDialog.InitialDir:= mountpoint;
  OpenDialog.Filter := 'Backup files (*.backup)|*.backup';
  if OpenDialog.Execute then
     begin
       VerifyFileName:= OpenDialog.Filename;
       getRestoreData(VerifyFileName,VSettings);
       with MainFM do
          begin
            VBox3Lbl.Caption:= '3 - Backup: ' + chopstr(VerifyFileName,45);
            VChkBox3.Enabled:= true;
            VChkBox3.Clear; //in case it has already been populated
            VDateLbl.Visible:= true;
            VDateLbl.Caption:= 'Backup Date:  ' + VSettings.BackDate + LF
                               + 'Time Now:       ' + FormatDateTime(FDateTimeStr,Now);
            VChkBox3.Items.Add('Select all partition images');
            for i:= 1 to VSettings.TotalParts do
                begin
                  VChkBox3.Items.Add(BuildPartStr(VSettings.NewParts[i]));
                  VChkBox3.Checked[i]:= false;
                  if length(VChkBox3.Items[i-1]) > maxlength then maxlength:= length(VChkBox3.Items[i-1]);
                end;
            //set horizontal scroll bar if needed
            if maxlength * 6 > VChkBox3.Width then VChkBox3.ScrollWidth:= VChkBox3.Width + 1
            else VChkBox3.ScrollWidth:= VChkBox3.Width;
            //if notes in backup make note button visible
            if VSettings.Note then
               begin
                 VNoteBtn.Enabled:= true;
                 VNoteBtn.Visible:= true;
               end;
         end
     end
  else //user has cancelled, have to clear box 3
     with MainFm do
          begin
            VChkBox2.CheckAll(cbUnchecked);
            VChkBox3.Clear;
            VChkBox3.Enabled:= false;
            VChkBox3.ScrollWidth:= ChkBoxWidth;
            VBox3Lbl.Caption:= '3 - Contents of Backup:';
          end;
  if ForceMount then
     begin
       //unmount the partition
     end;
  OpenDialog.Free;
end; //end procedure VGetFile
{-------------------------------------------------------------------------------
PROCEDURE VERIFYPARTS
Does the verify.

To do: Rsettings.ptable is redundant (and is set TRUE) it is simply recording whether the ptable is
selected in the settings tab, the ptable is always backed up and always restored if it has changed
(along with grub).

Used by: Unit mainform, VButtonCK

V1.0 - 14 Jan 2021
-------------------------------------------------------------------------------}
procedure VerifyParts;
var
  cmdstr : string; //verify command line
  StartTime, PstartTime : TDateTime;
  TotalBlocks : int64 = 0; //total blocks to restore
  PreviousBlocks : int64 = 0; //blocks already restored
  i, j : integer;
  Chkfilename, CheckedFname, BackFolder : string;
  Checkfiles, Verifytext, Verifytext2 : TStringList;
begin
  //debug
  if debug then DebugOut(true,'VerifyParts');
  //end debug
  StartTime:= Now;
  CurrentAction:= acVerify;
  //disable the main form
  MainFm.PageCtrl.Enabled:= false;
  ProgressFm:= TProgressFm.Create(MainFm);
  //Hide the add note button, not needed for verify
  ProgressFm.AddBtn.Enabled:= false;
  ProgressFm.AddBtn.Visible:= false;
  //make the log button visible
  ProgressFm.LogBtn.Visible:= true;;
  //enable the cancel button
  ProgressFm.CancelBtn.Enabled:= true;;
  ProgressFm.CancelBtn.Visible:= true;;
  ProgressFm.Show;
  //total blocks to verify
  for i:= 1 to VSettings.TotalParts do
      if MainFm.VChkBox3.Checked[i] then TotalBlocks+= VSettings.NewParts[i].blocks;
  ProgressFm.OverallProgBar.Max:= TotalBlocks;
  ProgressFm.PartListBox.Items.Add('Partition table - not selectable');
  //populate the list box with # parts
  for i:= 1 to VSettings.TotalParts do
      if MainFm.VChkBox3.Checked[i] then
         ProgressFm.PartListBox.Items.Add('Part ' + IntToStr(i) + ' - not started')
      else ProgressFm.PartListBox.Items.Add('Part ' + IntToStr(i) + ' - not selected');
  application.processmessages;
  //now run partclone
  for i:= 1 to MainFm.VChkBox3.Count - 1 do if MainFm.VChkBox3.Checked[i] then
     begin
       PstartTime:= Now;
       //set current progress bar
       ProgressFm.Caption:= 'Verifying Part ' + IntToStr(i)
                            + ' of ' + IntToStr(VSettings.TotalParts);
       //set the caption in the listbox
       ProgressFm.PartListBox.Items[i]:= 'Part '+ IntToStr(i) + ' - verifying';
       //build the partclone command line
       cmdstr:= PartCloneV(VerifyFilename,VSettings.NewParts[i].pnumber);
       writeln('verify command line =',cmdstr);
       //run partclone and update progress form
       RunPartClone(cmdstr,VSettings.NewParts[i].blocks,PreviousBlocks,TotalBlocks, StartTime);
       if CancelPartClone then
          begin
            CancelPartClone:= false;
            ProgressFm.Close;
            exit;
          end;
       //don't know what it outputs if the check fails so put in an else clause
       if ProgressFm.CurrentProgLbl.Caption = 'Cloned successfully' then //output from partclone
          begin
             ProgressFm.CurrentProgLbl.Caption:= 'Verified successfully';
             ProgressFm.PartListBox.Items[i]:= 'Part '+ IntToStr(i) + ' - verified '
                                               + FormatDateTime('hh:nn:ss',PstartTime - Now);
          end
       else ProgressFm.PartListBox.Items[i]:= 'Part '+ IntToStr(i) + ' - checked '
                                              + FormatDateTime('hh:nn:ss',PstartTime - Now);
       PreviousBlocks+= VSettings.NewParts[i].blocks;
       //now we have a chk file need to open and add to it
       Verifytext:= TStringList.create;
       //now we need the filename e.g. 20191110.sda5.img.chk
       ChkFilename:= ReplaceStr(VerifyFilename,'backup',VSettings.NewParts[i].pnumber) + '.img.chk';
       Verifytext.loadfromfile(ChkFilename); //looks like this overwrites the previous contents
       Verifytext.insert(0,(copy(ChkFilename,rpos('/',ChkFilename)+1,99)));
       Verifytext.insert(0,(FormatDateTime('dd mmm yyyy, hh:nn',Now)));
       Verifytext.Add('--------------------------------------------');
       Verifytext.savetofile(ChkFilename);
       Verifytext.free;
     end;
  BackFolder:= copy(VerifyFilename,0,rpos('/',VerifyFilename));
  Checkfiles:= TStringList.Create;
  //find all chk files, false is not search sub folders
  FindAllFiles(Checkfiles, BackFolder, '*.chk', false);
  Checkfiles.sort;
  //load all the chk files into a verify file
  if Checkfiles.count >= 1 then
     begin
       CheckedFname:= ReplaceStr(VerifyFilename,'.backup','.verify');
       Verifytext:= TStringList.create;
       Verifytext.loadfromfile(CheckFiles[0]);
       for i:= 1 to Checkfiles.count - 1 do
           begin
             Verifytext2:= TStringList.create;
             Verifytext2.loadfromfile(CheckFiles[i]);
             for j:= 0 to Verifytext2.count - 1 do Verifytext.add(Verifytext2[j]);
             Verifytext2.free;
           end;
       Verifytext.savetofile(CheckedFname);
       Verifytext.free;
     end;
  Checkfiles.free;
  //reset the progress form to indicate completion
  ProgressFm.Caption:= 'Verify complete';
  ProgressFm.PartcloneLbl.Caption:= '';
  ProgressFm.CancelBtn.Enabled:= false;
  ProgressFm.FinishBtn.Enabled:= true;
  //enable the view log button
  ProgressFm.LogBtn.Enabled:= true;;
end; //end procedure VerifyParts
{-------------------------------------------------------------------------------
PROCEDURE VERIFYLOG

Displays the log files after a verify.

Used by: Unit progform - ProgFmViewLog

V1.0 - 16 Jan 2021
-------------------------------------------------------------------------------}
procedure VerifyLog;
var
    VerifyNote : TNoteForm;
    VerifyFname : string;
begin
  //debug
  if debug then DebugOut(true,'VerifyLog');
  //end debug
  //have to show note and load it with the verify file
  //self gives an error nil does not????
  //need scroll bars - auto
  VerifyNote:= TNoteForm.create(ProgressFm);
  with VerifyNote do
       begin
         SaveButton.Enabled:= false;
         SaveButton.Visible:= false;
         CancelButton.Left:= 300;
         Caption:= 'Verify Log'; //caption for the title bar
       end;
  VerifyFname:= ReplaceStr(VerifyFilename,'backup','verify');
  VerifyNote.BackMemo.Lines.LoadFromFile(VerifyFname);
  VerifyNote.BackMemo.Font.Name:= 'Noto Sans Mono';
  VerifyNote.BackMemo.Font.Size:= 9;
  VerifyNote.show;
end;

end.

